from PyRlEnvs.Category import addToCategory
import numpy as np
from PyRlEnvs.FiniteDynamics import FiniteDynamics
from PyRlEnvs.utils.math import try2jit

LEFT = 0
RIGHT = 1

def _buildTransitionKernel(states: int):
    K = np.zeros((states, 2, states))

    mid = int(states // 2)
    last = states - 1

    # handle left side
    K[0, LEFT, mid] = 1
    K[0, RIGHT, 1] = 1

    # handle right side
    K[last, LEFT, last - 1] = 1
    K[last, RIGHT, mid] = 1

    # handle rest
    for s in range(1, last, 1):
        K[s, LEFT, s - 1] = 1
        K[s, RIGHT, s + 1] = 1

    return K

def _buildRewardKernel(states: int):
    Rs = np.zeros((states, 2, states))

    last = states - 1

    Rs[0, LEFT] = -1
    Rs[last, RIGHT] = 1

    return Rs

def _buildTerminationKernel(states: int):
    T = np.zeros((states, 2, states))

    last = states - 1

    T[0, LEFT] = 1
    T[last, RIGHT] = 1

    return T

def _buildStartStateDist(states: int):
    d0 = np.zeros(states)
    d0[int(states // 2)] = 1

    return d0

def buildRandomWalk(states: int):
    class RandomWalk(FiniteDynamics):
        K = _buildTransitionKernel(states)
        Rs = _buildRewardKernel(states)
        T = _buildTerminationKernel(states)
        d0 = _buildStartStateDist(states)

        num_states = states
        num_actions = 2

    return RandomWalk


# a default class, just for consistency
RandomWalk = buildRandomWalk(5)


# some utility functions to encode other important parts of the problem spec
# not necessarily environment specific, but this is as good a place as any to store them
@try2jit
def _normRows(m: np.ndarray):
    return (m.T / m.sum(axis=1)).T

"""
Feature representations used in
TODO: cite <Sutton et al. 2009>
"""
def invertedFeatures(n: int):
    # additive inverse of tabular (hence name)
    m = 1 - tabularFeatures(n)
    return _normRows(m)

@try2jit
def dependentFeatures(n: int):
    nfeats = int(np.floor(n / 2) + 1)
    m = np.zeros((n, nfeats))

    idx = 0
    for i in range(nfeats):
        m[idx, 0: i + 1] = 1
        idx += 1

    for i in range(nfeats - 1, 0, -1):
        m[idx, -i:] = 1
        idx += 1

    return _normRows(m)

def tabularFeatures(n: int):
    return np.eye(n)

addToCategory('ope', RandomWalk)
addToCategory('random-walk', RandomWalk)
addToCategory('finite-dynamics', RandomWalk)



from __future__ import annotations

import numpy as np
from abc import abstractmethod
from typing import Dict, Optional, Union

RNG = np.random.Generator

class Distribution:
    def __init__(self, rng: Optional[RNG] = None):
        self.rng = rng or np.random.default_rng()

    @abstractmethod
    def sample(self, rng: Optional[RNG] = None) -> float:
        pass

    def _rng(self, rng: Optional[RNG] = None):
        if rng:
            return rng

        if self.rng:
            return self.rng

        return np.random

    def __add__(self, other: Union[float, Distribution]):
        if isinstance(other, float):
            other = DeltaDist(other, self.rng)

        return AddedDist(self, other)

    def __mul__(self, other: Union[float, Distribution]):
        if isinstance(other, float):
            other = DeltaDist(other, self.rng)

        return MultipliedDist(self, other)

    def __rmul__(self, other: Union[float, Distribution]):
        return self.__mul__(other)

class AddedDist(Distribution):
    def __init__(self, ldist: Distribution, rdist: Distribution):
        self.ldist = ldist
        self.rdist = rdist

    def sample(self, rng: Optional[RNG] = None):
        return self.ldist.sample(rng) + self.rdist.sample(rng)

class MultipliedDist(Distribution):
    def __init__(self, ldist: Distribution, rdist: Distribution):
        self.ldist = ldist
        self.rdist = rdist

    def sample(self, rng: Optional[RNG] = None):
        return self.ldist.sample(rng) * self.rdist.sample(rng)

class DeltaDist(Distribution):
    def __init__(self, val: float, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.value = val

    def sample(self, rng: Optional[RNG] = None) -> float:
        return self.value

class Gaussian(Distribution):
    def __init__(self, mean: float, stddev: float, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.mean = mean
        self.stddev = stddev

    def sample(self, rng: Optional[RNG] = None):
        return self._rng(rng).normal(self.mean, self.stddev)

class Uniform(Distribution):
    def __init__(self, mi: float, ma: float, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.mi = mi
        self.ma = ma

    def sample(self, rng: Optional[RNG] = None):
        return self._rng(rng).uniform(self.mi, self.ma)


class Gamma(Distribution):
    def __init__(self, shape: float, scale: float, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.shape = shape
        self.scale = scale

    def sample(self, rng: Optional[RNG] = None):
        return self._rng(rng).gamma(self.shape, self.scale)

class ClippedGaussian(Gaussian):
    def __init__(self, mean: float, stddev: float, mi: float, ma: Optional[float] = None, rng: Optional[RNG] = None):
        super().__init__(mean, stddev, rng)

        self.mi = mi
        self.ma = ma

    def sample(self, rng: Optional[RNG] = None):
        x = super().sample(rng)
        return np.clip(x, self.mi, self.ma)

def sampleChildren(d: Dict[str, Distribution], rng: Optional[RNG] = None):
    out = {}
    for k in d:
        out[k] = d[k].sample(rng)

    return out



from PyRlEnvs.Category import addToCategory
import numpy as np
from PyRlEnvs.domains.GridWorld.Elements import Element, StartState, GoalState
from PyRlEnvs.domains.GridWorld.utils import Coords, predecessors
from PyRlEnvs.domains.GridWorld import GridWorldBuilder

class Cliff(Element):
    def __init__(self, loc: Coords):
        super().__init__('Cliff')

        self.loc = loc

    def trigger(self, s: int, a: int, sp: int) -> bool:
        target = self.getState(self.loc)
        return sp == target

    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        target = self.getState(self.loc)
        starts = np.argwhere(d0 > 0)

        # instead of transitioning into the cliff
        # we need to transition back to the start
        K[:, :, target] = 0
        R[:, :, target] = 0
        T[:, :, target] = 0
        d0[target] = 0

        for (s, a) in predecessors(target, self.shape):
            # reminder that we *don't* ever actually transition into the cliff
            # we end up back in the start state instead
            R[s, a, starts] = -100

            # probably there is only one start state
            # but in case we ever define multiple, then uniform random
            # chance to transition to any of them
            for start in starts:
                K[s, a, start] = 1.0 / len(starts)


def build(shape: Coords = (12, 4)):
    cliffWorldBuilder = GridWorldBuilder(shape)

    # start in bottom left
    cliffWorldBuilder.addElement(StartState((0, 0)))

    # end in bottom right without additional reward
    cliffWorldBuilder.addElement(GoalState((shape[0] - 1, 0), -1))

    # with a cliff in every state in-between
    cliffWorldBuilder.addElements([ Cliff((x, 0)) for x in range(1, shape[0] - 1) ])

    cliffWorldBuilder.costToGoal = True

    return cliffWorldBuilder.build()

CliffWorld = build()

addToCategory('gridworld', CliffWorld)
addToCategory('finite-dynamics', CliffWorld)
addToCategory('sutton-barto', CliffWorld)



"""
Simulation of HIV Treatment. The aim is to find an optimal drug schedule.

The state contains concentrations of 6 different cells:
* T1: non-infected CD4+ T-lymphocytes [cells / ml]
* T1*:    infected CD4+ T-lymphocytes [cells / ml]
* T2: non-infected macrophages [cells / ml]
* T2*:    infected macrophages [cells / ml]
* V: number of free HI viruses [copies / ml]
* E: number of cytotoxic T-lymphocytes [cells / ml]

The therapy consists of 2 drugs
(reverse transcriptase inhibitor [RTI] and protease inhibitor [PI]) which
are activated or not. The action space contains therefore of 4 actions:
* *0*: none active
* *1*: RTI active
* *2*: PI active
* *3*: RTI and PI active

Ernst, D., Stan, G., Gonc, J. & Wehenkel, L.
Clinical data based optimal STI strategies for HIV:
A reinforcement learning approach
In Proceedings of the 45th IEEE Conference on Decision and Control (2006).
"""

import numpy as np
from PyRlEnvs.Category import addToCategory
from PyRlEnvs.utils.math import clipEach, try2jit
from PyRlEnvs.utils.RandomVariables import DeterministicRandomVariable
from PyRlEnvs.BaseEnvironment import BaseEnvironment
from scipy.integrate import solve_ivp

@try2jit
def _dsdt(t: float, sa: np.ndarray):
    # putting these in the method lets numba treat them as constants
    # on the other hand it prevents us from modifying them in derivative classes
    lambda1 = 1e4
    lambda2 = 31.98
    d1 = 0.01
    d2 = 0.01
    f = .34
    k1 = 8e-7
    k2 = 1e-4
    delta = .7
    m1 = 1e-5
    m2 = 1e-5
    nt = 100
    c = 13.
    rho1 = 1.
    rho2 = 1.
    lambdaE = 1
    be = 0.3
    kb = 100
    de = 0.25
    kd = 500
    deltaE = 0.1

    t1, t2, t1s, t2s, v, e, eps1, eps2 = sa

    tmp1 = (1. - eps1) * k1 * v * t1
    tmp2 = (1. - f * eps1) * k2 * v * t2

    dt1 = lambda1 - d1 * t1 - tmp1
    dt2 = lambda2 - d2 * t2 - tmp2
    dt1s = tmp1 - delta * t1s - m1 * e * t1s
    dt2s = tmp2 - delta * t2s - m2 * e * t2s

    dv = (1 - eps2) * nt * delta * (t1s - t2s) - c * v - ((1 - eps1) * rho1 * k1 * t1 + (1 - f * eps1) * rho2 * k2 * t2) * v
    de = lambdaE + be * (t1s - t2s) / (t1s + t2s + kb) * e - de * (t1s - t2s) / (t1s + t2s + kd) * e - deltaE * e

    return np.array([dt1, dt2, dt1s, dt2s, dv, de, 0, 0])

def _nextState(s: np.ndarray, a: int, dt: float, effects: np.ndarray) -> np.ndarray:
    eps1, eps2 = effects[a]
    sa = np.append(s, [eps1, eps2])
    spa = solve_ivp(_dsdt, [0, dt], sa).y.T

    # only need the last result of the integration
    spa = spa[-1]
    sp = spa[:-2]

    sp = clipEach(sp, 1e-8, 1e8)

    return sp

@try2jit
def _transform(s: np.ndarray) -> np.ndarray:
    return np.log10(s)

class HIVTreatment(BaseEnvironment):
    dt = 5

    actionEffects = np.array([[0., 0], [.7, 0], [0, .3], [.7, .3]])

    @classmethod
    def nextStates(cls, s: np.ndarray, a: int):
        sp = _nextState(s, a, cls.dt, cls.actionEffects)
        return DeterministicRandomVariable(_transform(sp))

    @classmethod
    def actions(cls, s: np.ndarray):
        return [0, 1, 2, 3]

    @classmethod
    def reward(cls, s: np.ndarray, a: int, sp: np.ndarray):
        _, _, _, _, v, e = sp

        eps1, eps2 = cls.actionEffects[a]
        return -0.1 * v - 2e4 * eps1**2 - 2e3 * eps2**2 + 1e3 * e

    @classmethod
    def terminal(cls, s: np.ndarray, a: int, sp: np.ndarray):
        return False

    def __init__(self, random_start: float = 0, seed: int = 0):
        super().__init__(seed)
        self.random_start = random_start
        self.start_rng = np.random.default_rng(seed)

        self._state: np.ndarray = np.zeros(6)

    def start(self):
        # perturb the start state multiplicatively by a normal random amount for each component of system
        eps = 1
        if self.random_start > 0:
            eps = self.start_rng.normal(1, self.random_start, size=6)

        # start in non-healthy stable state
        start = np.array([163573, 5., 11945, 46, 63919, 24]) * eps
        self._state = start

        return _transform(start)

    def step(self, action: int):
        sp = _nextState(self._state, action, self.dt, self.actionEffects)
        r = HIVTreatment.reward(self._state, action, sp)
        t = HIVTreatment.terminal(self._state, action, sp)

        self._state = sp

        return (r, _transform(sp), t)

    def setState(self, state: np.ndarray):
        self._state = state.copy()

    def copy(self):
        m = HIVTreatment(self._seed)
        m._state = self._state.copy()
        return m

addToCategory('classic-control', HIVTreatment)



"""
TODO:
From <cite Boyan>
"""

from PyRlEnvs.utils.math import immutable
from PyRlEnvs.Category import addToCategory
import numpy as np
from PyRlEnvs.FiniteDynamics import FiniteDynamics

RIGHT = 0
SKIP = 1


def _buildTransitionKernel():
    K = np.zeros((13, 2, 13))
    for i in range(11):
        K[i, RIGHT, i + 1] = 1
        K[i, SKIP, i + 2] = 1

    K[11, RIGHT, 12] = 1
    K[12, RIGHT, 0] = 1

    return K

def _buildRewardKernel():
    R = np.zeros((13, 2, 13))
    R[:11, :] = -3
    R[11, RIGHT, 12] = -2

    return R

def _buildTerminationKernel():
    T = np.zeros((13, 2, 13))
    T[12, 0, 0] = 1

    return T

class BoyanChain(FiniteDynamics):
    num_states = 13
    num_actions = 2

    K = immutable(_buildTransitionKernel())
    Rs = immutable(_buildRewardKernel())

    T = immutable(_buildTerminationKernel())
    d0 = immutable(np.array([1.] + [0.] * 12))

# some utility functions to encode other important parts of the problem spec
# not necessarily environment specific, but this is as good a place as any to store them
def behaviorPolicy(s: int):
    if s <= 10:
        return np.array([0.5, 0.5])

    return np.array([1.0, 0])

def representationMatrix():
    return np.array([
        [1,    0,    0,    0   ],  # noqa: E241
        [0.75, 0.25, 0,    0   ],  # noqa: E241
        [0.5,  0.5,  0,    0   ],  # noqa: E241
        [0.25, 0.75, 0,    0   ],  # noqa: E241
        [0,    1,    0,    0   ],  # noqa: E241
        [0,    0.75, 0.25, 0   ],  # noqa: E241
        [0,    0.5,  0.5,  0   ],  # noqa: E241
        [0,    0.25, 0.75, 0   ],  # noqa: E241
        [0,    0,    1,    0   ],  # noqa: E241
        [0,    0,    0.75, 0.25],  # noqa: E241
        [0,    0,    0.5,  0.5 ],  # noqa: E241
        [0,    0,    0.25, 0.75],  # noqa: E241
        [0,    0,    0,    1   ],  # noqa: E241
    ])

addToCategory('ope', BoyanChain)
addToCategory('random-walk', BoyanChain)
addToCategory('finite-dynamics', BoyanChain)



from typing import Any, Sequence, Set, TypeVar, Union
import numpy as np
from numba import njit

T = TypeVar('T')
NpArray = Union[Sequence[float], np.ndarray]

# we can speed this up if we abstract away from the rng
@njit(cache=True)
def _sample(arr: NpArray, r: float) -> int:
    s = 0
    for i, p in enumerate(arr):
        s += p
        if s > r or s == 1:
            return i

    # worst case if we run into floating point error, just return the last element
    # we should never get here
    return len(arr) - 1

# way faster than np.random.choice
# arr is an array of probabilities, should sum to 1
def sample(arr: NpArray, rng: Any = np.random) -> int:
    # if we can avoid incrementing the rng, do so
    if len(arr) == 1:
        return 0

    r = rng.random()
    return _sample(arr, r)

# also much faster than np.random.choice
# choose an element from a list with uniform random probability
def choice(arr: Union[Sequence[T], Set[T]], rng: Any = np.random) -> T:
    idx = rng.randint(0, len(arr))

    return list(arr)[idx]



import numpy as np
from abc import abstractmethod
from .utils import Coords, getState, predecessors

class Element:
    def __init__(self, name: str = ''):
        self.name = name
        self.shape: Coords = (0, 0)

    # deferred initialization for when we know more about our gridworld
    def init(self, shape: Coords):
        self.shape = shape

    def getState(self, coords: Coords) -> int:
        return getState(coords, self.shape)

    # Used for printing to console. Can be safely ignored without impacting behavior
    def trigger(self, s: int, a: int, sp: int) -> bool:
        return False

    @abstractmethod
    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        pass

class GoalState(Element):
    def __init__(self, loc: Coords, reward: float):
        super().__init__('Goal')
        self.loc = loc
        self.reward = reward

    def init(self, shape: Coords):
        self.shape = shape

        self.sp = self.getState(self.loc)

    def trigger(self, s: int, a: int, sp: int):
        return sp == self.sp

    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        for (s, a) in predecessors(self.sp, self.shape):
            R[s, a, self.sp] = self.reward
            T[s, a, self.sp] = 1

class StartState(Element):
    def __init__(self, loc: Coords, weight: float = 1):
        super().__init__('Start')
        self.loc = loc
        self.weight = weight

    def init(self, shape: Coords):
        self.shape = shape

        self.s = self.getState(self.loc)

    def trigger(self, s: int, a: int, sp: int):
        return s == self.s

    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        d0[self.s] = self.weight

class WallState(Element):
    def __init__(self, loc: Coords):
        super().__init__('Wall')
        self.loc = loc

    def init(self, shape: Coords):
        self.shape = shape

        self.sp = self.getState(self.loc)

    def trigger(self, s: int, a: int, sp: int):
        return sp == self.sp

    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        sp = self.sp

        for (s, a) in predecessors(sp, self.shape):
            d0[sp] = 0
            K[s, a, s] = 1
            R[s, a, s] = R[s, a, sp]
            K[:, :, sp] = 0
            R[:, :, sp] = 0



from typing import Dict, Iterator, List, Optional, NamedTuple, Type
from PyRlEnvs.BaseEnvironment import BaseEnvironment

class EnvironmentMetadata(NamedTuple):
    name: str
    Env: Type[BaseEnvironment]
    description: Optional[str] = None


# TODO: generate documentation based off of this information
# TODO: generate experiment descriptions based off of this (not in this repo, but this is best place to store this todo right now)
class Category:
    def __init__(self, name: str, description: Optional[str] = None):
        self.name = name
        self.description = description

        self.environments: List[EnvironmentMetadata] = []

    def addEnvironment(self, name: str, Env: Type[BaseEnvironment], description: Optional[str] = None):
        metadata = EnvironmentMetadata(name, Env, description)
        self.environments.append(metadata)

        return metadata

    def __len__(self) -> int:
        return len(self.environments)

    def __iter__(self) -> Iterator[EnvironmentMetadata]:
        return self.environments.__iter__()

    def __getitem__(self, idx: int) -> EnvironmentMetadata:
        return self.environments.__getitem__(idx)

_categories: Dict[str, Category] = {}

def addToCategory(category: str, Env: Type[BaseEnvironment], description: Optional[str] = None):
    cat = _categories.get(category)

    if cat is None:
        cat = Category(category)

    cat.addEnvironment(Env.__name__, Env, description)

    return cat

def createCategory(category: str, description: Optional[str] = None):
    if category in _categories:
        print('Overwritting this category ', category)

    _categories[category] = Category(category, description)

    return _categories[category]

def getCategory(category: str):
    cat = _categories.get(category)

    if cat is None:
        raise Exception("Nope, this category doesn't exist", category)

    return cat



from functools import partial
import numpy as np
import numpy.typing as npt
from numpy.random import default_rng
from PyRlEnvs.BaseEnvironment import BaseEnvironment

from numba import njit

UP = 0
DOWN = 1
RIGHT = 2
LEFT = 3

@njit(cache=True)
def l2_dist(x: np.ndarray, y: np.ndarray):
    return np.sqrt(np.power(x - y, 2).sum())

@njit(cache=True)
def puddleDist(
        # parameters of the puddle
        head: np.ndarray,
        tail: np.ndarray,
        radius: float,
        length: float,
        axis: int,

        # agent parameters
        pos: np.ndarray,
    ) -> float:
    u = (pos - tail)[axis] / length

    dist = 0.0

    if u < 0.0:
        dist = l2_dist(tail, pos)
    elif u > 1.0:
        dist = l2_dist(head, pos)
    else:
        pos_ = tail + u * (head - tail)
        dist = l2_dist(pos, pos_)

    if dist < radius:
        return (radius - dist)

    return 0.

def buildPuddle(
        head: npt.ArrayLike,
        tail: npt.ArrayLike,
        radius: float,
        length: float,
        axis: int,
    ):

    head_ = np.asarray(head)
    tail_ = np.asarray(tail)

    return partial(puddleDist, head_, tail_, radius, length, axis)

class PuddleWorld(BaseEnvironment):
    def __init__(self, seed: int = 0):
        super().__init__(seed)

        # two rngs. One gives a sequence of perturbations per step
        # the other gives a sequence of start states
        self.rng = default_rng(seed)
        self.start_rng = default_rng(seed)

        self.state = np.zeros(2)

        # Add the puddles
        self.puddle1 = buildPuddle(
            head=(0.45, 0.75),
            tail=(0.1, 0.75),
            radius=0.1,
            length=0.35,
            axis=0,
        )

        self.puddle2 = buildPuddle(
            head=(0.45, 0.8),
            tail=(0.45, 0.4),
            radius=0.1,
            length=0.4,
            axis=1,
        )

        self.goal_dimension = 0.05
        self.def_displacement = 0.05

        self.sigma = 0.01

        self.goal = 1 - self.goal_dimension

    @classmethod
    def actions(cls, s):
        return [UP, DOWN, RIGHT, LEFT]

    def start(self):
        # start in a region around (0.05, 0.05) randomly
        # enforce boundaries of 1x1 square
        self.state = self.start_rng.normal(0.05, 0.1, size=2)
        self.state = np.clip(self.state, 0, 1)
        return np.copy(self.state)

    def _terminal(self):
        return np.all(self.state >= self.goal)

    def _reward(self, state, terminal):
        if terminal:
            return -1.

        reward = -1 + self.puddle1(state) * -400 + self.puddle2(state) * -400
        return reward

    def step(self, a):
        s = self.state

        n = self.rng.normal(scale=self.sigma)

        if a == UP: #up
            s[1] += (self.def_displacement+n)
        elif a == DOWN: #down
            s[1] -= (self.def_displacement+n)
        elif a == RIGHT: #right
            s[0] += (self.def_displacement+n)
        elif a == LEFT: #left
            s[0] -= (self.def_displacement+n)
        else:
            raise Exception()

        if s[0] > 1:
            s[0] = 1
        elif s[0] < 0:
            s[0] = 0

        if s[1] > 1:
            s[1] = 1
        elif s[1] < 0:
            s[1] = 0

        self.state = s

        t = self._terminal()
        r = self._reward(s, t)

        return (r, np.copy(self.state), t)



import numpy as np
from functools import partial

from PyRlEnvs.Category import addToCategory
from PyRlEnvs.utils.distributions import ClippedGaussian, DeltaDist, Gamma, Gaussian, Uniform, sampleChildren
from PyRlEnvs.BaseEnvironment import BaseEnvironment
from PyRlEnvs.utils.math import try2jit
from PyRlEnvs.utils.numerical import euler

@try2jit
def _dsdt(g: float, l: float, masspole: float, masscart: float, sa: np.ndarray, t: float):
    polemass_length = masspole * l
    total_mass = masspole + masscart

    x, dx, theta, dtheta, force = sa

    sinTheta = np.sin(theta)
    cosTheta = np.cos(theta)

    term1: float = (force + polemass_length * dtheta**2 * sinTheta) / total_mass
    ddtheta: float = (g * sinTheta - cosTheta * term1) / (l * (4 / 3 - masspole * cosTheta**2 / total_mass))
    ddx: float = term1 - polemass_length * ddtheta * cosTheta / total_mass

    return np.array([dx, ddx, dtheta, ddtheta, 0])

@try2jit
def _isTerminal(s: np.ndarray) -> bool:
    x, _, theta, _ = s
    theta_thresh = 12 * 2 * np.pi / 360
    return x < -2.4 or x > 2.4 or theta < -theta_thresh or theta > theta_thresh

class Cartpole(BaseEnvironment):
    physical_constants = {
        # physics of world / cart / pole
        'gravity': 9.8,
        'pole_length': 0.5,
        'pole_mass': 0.1,
        'cart_mass': 1.0,
    }

    per_step_constants = {
        # for integration fidelity
        'dt': DeltaDist(0.02),

        # controller force
        'force': DeltaDist(10),
    }

    randomized_constants = {
        'gravity': ClippedGaussian(mean=9.8, stddev=2.0, mi=5.0, ma=13.8),
        'pole_length': Uniform(mi=0.3, ma=0.7),
        'pole_mass': Uniform(mi=0.05, ma=0.15),
        'cart_mass': Uniform(mi=0.6, ma=1.4),
    }

    per_step_random_constants = {
        # make time sampled from a normal distribution (clipped to ensure non-negative) with a long-tailed nuisance
        # distribution to simulate random delays or interference
        'dt': 0.95 * ClippedGaussian(mean=0.02, stddev=0.01, mi=0.015, ma=None) + 0.05 * Gamma(shape=0.01, scale=2.0),

        'force': Gaussian(mean=10, stddev=1.0),
    }

    def __init__(self, randomize: bool = False, seed: int = 0):
        super().__init__(seed)
        self.randomize = randomize
        self._state = np.zeros(4)

        self.start_rng = np.random.default_rng(seed)

        if randomize:
            self.physical_constants = sampleChildren(self.randomized_constants, self.rng)
            self.per_step_constants = self.per_step_random_constants

        self._dsdt = partial(_dsdt,
            self.physical_constants['gravity'],
            self.physical_constants['pole_length'],
            self.physical_constants['pole_mass'],
            self.physical_constants['cart_mass'],
        )

    # -------------------------
    # -- Dynamics equations --
    # -------------------------

    def nextState(self, s: np.ndarray, a: int):
        # get per-step constants
        dt = self.per_step_constants['dt'].sample(self.rng)
        force = self.per_step_constants['force'].sample(self.rng)

        force = force if a == 1 else -force

        sa = np.append(s, force)
        spa = euler(self._dsdt, sa, np.array([0, dt]))

        # only need the last result of the integration
        spa = spa[-1]
        sp = spa[:-1]

        return sp

    def actions(self, s: np.ndarray):
        return [0, 1]

    def reward(self, s: np.ndarray, a: int, sp: np.ndarray):
        return 1.0

    def terminal(self, s: np.ndarray, a: int, sp: np.ndarray):
        return _isTerminal(sp)

    # ------------------------
    # -- Stateful functions --
    # ------------------------

    def start(self):
        start = self.start_rng.uniform(-0.05, 0.05, size=4)
        self._state = start
        return start

    def step(self, action: int):
        sp = self.nextState(self._state, action)
        r = self.reward(self._state, action, sp)
        t = self.terminal(self._state, action, sp)

        self._state = sp

        return (r, sp, t)

    def setState(self, state: np.ndarray):
        self._state = state.copy()

    def copy(self):
        m = Cartpole(randomize=self.randomize, seed=self._seed)
        m._state = self._state.copy()
        m.physical_constants = self.physical_constants
        m.per_step_constants = self.per_step_constants

        # because we are changing the physics _after_ the derivatives are being precomputed
        # we need to replace the derivatives with the correct eqns for these constants
        m._dsdt = self._dsdt

        return m

class StochasticCartpole(Cartpole):
    def __init__(self, seed: int = 0):
        super().__init__(randomize=True, seed=seed)

class ContinuousActionCartpole(Cartpole):
    def nextState(self, s: np.ndarray, force: float):
        # get per-step constants
        dt = self.per_step_constants['dt'].sample(self.rng)

        force = np.clip(force, -12, 12)
        sa = np.append(s, force)
        spa = euler(self._dsdt, sa, np.array([0, dt]))

        # only need the last result of the integration
        spa = spa[-1]
        sp = spa[:-1]

        return sp

addToCategory('classic-control', Cartpole)
addToCategory('stochastic', StochasticCartpole)



from __future__ import annotations
import numpy as np
import RlGlue
from abc import abstractmethod
from typing import Any


class BaseEnvironment(RlGlue.BaseEnvironment):
    def __init__(self, seed: int = 0):
        self._seed = seed
        self.rng = np.random.default_rng(seed)

    @abstractmethod
    def setState(self, state: Any):
        pass

    @abstractmethod
    def copy(self) -> BaseEnvironment:
        pass



from typing import Any, Sequence, Tuple

from PyRlEnvs.utils.math import try2FastJit, clipInt

Coords = Tuple[int, int]

@try2FastJit
def getState(coords: Coords, shape: Coords) -> int:
    x, y = coords
    mx, my = shape

    s = clipInt(y, 0, my) + clipInt(x, 0, mx) * my
    return int(s)

@try2FastJit
def getCoords(state: int, shape: Coords) -> Coords:
    x = int(state // shape[1])
    y = int(state % shape[1])

    return (x, y)

def findFirstTrigger(arr: Sequence[Any], s: int, a: int, sp: int):
    for element in arr:
        if element.trigger(s, a, sp) and element.name:
            return element

    return None

@try2FastJit
def predecessor(sp: int, action: int, shape: Coords):
    x, y = getCoords(sp, shape)

    # need to trick numba into knowing the type of this list
    empty = [i for i in range(0)]

    # UP
    if action == 0:
        if y == 0:
            return empty

        s = getState((x, y - 1), shape)

        if y == shape[1] - 1:
            return [sp, s]

        return [s]

    # RIGHT
    elif action == 1:
        if x == 0:
            return empty

        s = getState((x - 1, y), shape)

        if x == shape[0] - 1:
            return [sp, s]

        return [s]

    # DOWN
    elif action == 2:
        if y == shape[1] - 1:
            return empty

        s = getState((x, y + 1), shape)

        if y == 0:
            return [sp, s]

        return [s]

    # LEFT
    else:
        if x == shape[0] - 1:
            return empty

        s = getState((x + 1, y), shape)

        if x == 0:
            return [sp, s]

        return [s]

def predecessors(sp: int, shape: Coords):
    for a in range(4):
        s = predecessor(sp, a, shape)
        for _s in s:
            yield _s, a



import numpy as np
from abc import abstractmethod
from typing import Any, Callable, Generic, Optional, Sequence, TypeVar, Union, cast
from PyRlEnvs.utils.random import NpArray, sample

T = TypeVar('T')

FloatArray = Union[Sequence[float], np.ndarray]
RNG = np.random.Generator

class RandomVariable(Generic[T]):
    def __init__(self, rng: Optional[RNG] = None):
        if rng is None:
            rng = np.random.default_rng()

        self.rng: RNG = rng

    def _rng(self, rng: Optional[RNG] = None) -> RNG:
        if rng is not None:
            return rng

        return self.rng

    @abstractmethod
    def sample(self, rng: Optional[RNG] = None) -> T:
        pass

class DeterministicRandomVariable(RandomVariable[T]):
    def __init__(self, val: T, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.val = val

    def sample(self, rng: Optional[RNG] = None):
        return self.val

    def __iter__(self):
        yield self.val

class DiscreteRandomVariable(RandomVariable[T]):
    def __init__(self, vals: Sequence[T], probs: FloatArray, rng: Optional[RNG] = None):
        super().__init__(rng)

        self.vals = vals
        self.probs = probs

    def sample(self, rng: Optional[RNG] = None) -> T:
        rng = self._rng(rng)
        idx = sample(self.probs, rng)
        return self.vals[idx]

    def __iter__(self):
        return self.vals.__iter__()

    @staticmethod
    def fromProbs(probs: NpArray, rng: Optional[RNG] = None):
        vals = cast(Sequence[int], np.arange(len(probs)))
        return DiscreteRandomVariable[int](vals, probs, rng)

    @staticmethod
    def fromUniform(vals: Sequence[T], rng: Optional[RNG] = None):
        probs: np.ndarray = np.ones(len(vals)) / len(vals)
        return DiscreteRandomVariable[T](vals, probs, rng)

class InferedRandomVariable(RandomVariable[T]):
    def __init__(self, sampling_func: Callable[..., T], rng: Optional[RNG] = None):
        super().__init__(rng)

        self.sampling_func = sampling_func

    def sample(self, *args: Any, rng: Optional[RNG] = None, **kwargs: Any):
        rng = self._rng(rng)
        return self.sampling_func(rng, *args, **kwargs)



from PyRlEnvs.utils.RandomVariables import DiscreteRandomVariable
from typing import Callable, Optional, Sequence, cast
import numpy as np
from numba import njit
from PyRlEnvs.BaseEnvironment import BaseEnvironment
from PyRlEnvs.utils.random import sample

@njit(cache=True)
def _actions(K: np.ndarray, s: int):
    return np.unique(np.where(K[s] != 0)[0])

@njit(cache=True)
def _nextStates(K: np.ndarray, s: int, a: int):
    return np.where(K[s, a] > 0)[0]

@njit(cache=True)
def _transitionMatrix(K: np.ndarray, T: np.ndarray, pi: np.ndarray, gamma: float):
    states = pi.shape[0]
    P = np.zeros((states, states))

    g: np.ndarray = (1 - T) * gamma
    if gamma < 0:
        g = np.ones_like(T)

    for s in range(states):
        for sp in range(states):
            P[s, sp] = np.sum(K[s, :, sp] * pi[s] * g[s, :, sp])

    return P

@njit(cache=True)
def _averageReward(K: np.ndarray, Rs: np.ndarray, pi: np.ndarray):
    states = pi.shape[0]

    R = np.zeros(states)
    for s in range(states):
        for sp in range(states):
            R[s] += np.sum(K[s, :, sp] * Rs[s, :, sp] * pi[s])

    return R

@njit(cache=True)
def _stateDistribution(P: np.ndarray):
    return np.linalg.matrix_power(P, 1024).sum(axis=0) / P.shape[0]

class FiniteDynamics(BaseEnvironment):
    # start state dist
    d0 = np.zeros(0)

    # transition kernel
    # shape: (states, actions, states)
    K = np.zeros(())

    # reward kernel
    # shape: (states, actions, states)
    Rs = np.zeros(())

    # termination kernel
    # shape: (states, actions, states)
    T = np.zeros(())

    num_states = 0
    num_actions = 0

    @classmethod
    def actions(cls, s: int):
        # the available actions are any action where the kernel is non-zero for the given state
        return _actions(cls.K, s)

    @classmethod
    def nextStates(cls, s: int, a: int):
        sp = _nextStates(cls.K, s, a)
        sp = cast(Sequence[int], sp)
        return DiscreteRandomVariable(sp, cls.K[s, a, sp])

    @classmethod
    def reward(cls, s: int, a: int, sp: int) -> float:
        return cls.Rs[s, a, sp]

    @classmethod
    def terminal(cls, s: int, a: int, sp: int):
        return bool(cls.T[s, a, sp])

    @classmethod
    def constructTransitionMatrix(cls, policy: Callable[[int], np.ndarray], gamma: Optional[float] = None):
        if gamma is None:
            gamma = -1

        states = cls.num_states
        pi = np.array([ policy(s) for s in range(states) ])

        return _transitionMatrix(cls.K, cls.T, pi, gamma)

    @classmethod
    def constructRewardVector(cls, policy: Callable[[int], np.ndarray]):
        states = cls.num_states
        pi = np.array([ policy(s) for s in range(states) ])

        return _averageReward(cls.K, cls.Rs, pi)

    @classmethod
    def computeStateDistribution(cls, policy: Callable[[int], np.ndarray]):
        P = cls.constructTransitionMatrix(policy)
        return _stateDistribution(P)

    def __init__(self, seed: int = 0):
        super().__init__()

        self.rng = np.random.default_rng(seed)
        self.state: int = 0

    def start(self):
        self.state = sample(self.d0, self.rng)
        return self.state

    def step(self, action: int):
        sp = self.nextStates(self.state, action).sample(self.rng)
        r = self.reward(self.state, action, sp)
        t = self.terminal(self.state, action, sp)

        self.state = sp

        return (r, self.state, t)

    def setState(self, state: int):
        self.state = state

    def copy(self):
        c = self.__class__(self._seed)
        c.state = self.state

        return c



import logging
import numpy as np
from typing import Any, Callable, TypeVar

# -----------------
# -- Numba stuff --
# -----------------

_useFastMath = False
_haveWarned = False
def useFastMath(warn: bool = True):
    global _useFastMath, _haveWarned
    _useFastMath = True

    if not _haveWarned and warn:
        _haveWarned = True
        logging.getLogger('PyRlEnvs').warning('Enabling fastmath can result in nondeterministic accumulation of floating point errors')

F = TypeVar('F', bound=Callable[..., Any])
def try2jit(f: F) -> F:
    try:
        from numba import njit
        return njit(f, cache=True, nogil=True, fastmath=_useFastMath)

    except Exception:
        return f

# many of our functions operate on scalars or have operations which do not
# accumulate floating point errors. These can safely use fastmath
# without incurring nondeterminism.
def try2FastJit(f: F) -> F:
    try:
        from numba import njit
        return njit(f, cache=True, nogil=True, fastmath=True)

    except Exception:
        return f

# ----------------
# -- Math stuff --
# ----------------

@try2FastJit
def wrap(x: float, mi: float, ma: float):
    d = ma - mi

    while x > ma:
        x = x - d

    while x < mi:
        x = x + d

    return x

@try2FastJit
def clipEach(x: np.ndarray, mi: float, ma: float):
    out = np.zeros_like(x)
    for i in range(x.shape[0]):
        out[i] = clip(x[i], mi, ma)

    return out

@try2FastJit
def clip(x: float, mi: float, ma: float):
    if x > ma:
        return ma

    if x < mi:
        return mi

    return x

@try2FastJit
def clipInt(x: int, mi: int, ma: int):
    if x >= ma:
        return ma - 1

    if x < mi:
        return mi

    return x

def immutable(arr: np.ndarray):
    arr.setflags(write=False)
    return arr



"""doc
TODO: find original citation
"""

import numpy as np
from functools import partial
from PyRlEnvs.utils.distributions import ClippedGaussian, DeltaDist, Gamma, Gaussian, Uniform, sampleChildren
from PyRlEnvs.Category import addToCategory
from PyRlEnvs.utils.math import clip, try2jit, wrap
from PyRlEnvs.utils.numerical import rungeKutta
from PyRlEnvs.BaseEnvironment import BaseEnvironment

@try2jit
def _dsdt(l1: float, m1: float, m2: float, com1: float, com2: float, g: float, sa: np.ndarray, t: float):
    moi = 1.    # moment of inertia

    a = sa[-1]
    theta1, theta2, dtheta1, dtheta2 = sa[:-1]

    d1: float = m1 * com1**2 + m2 * (l1**2 + com2**2 + 2 * l1 * com2 * np.cos(theta2)) + 2 * moi
    d2: float = m2 * (com2**2 + l1 * com2 * np.cos(theta2)) + moi

    phi2: float = m2 * com2 * g * np.cos(theta1 + theta2 - (np.pi / 2))
    phi1: float = -m2 * l1 * com2 * dtheta2**2 * np.sin(theta2) - 2 * m2 * l1 * com2 * dtheta2 * dtheta1 * np.sin(theta2) + (m1 * com1 + m2 * l1) * g * np.cos(theta1 - (np.pi / 2)) + phi2

    ddtheta2 = (a + d2 / d1 * phi1 - m2 * l1 * com2 * dtheta1**2 * np.sin(theta2) - phi2) / (m2 * com2**2 + moi - (d2**2 / d1))
    ddtheta1 = -(d2 * ddtheta2 + phi1) / d1

    return np.array([dtheta1, dtheta2, ddtheta1, ddtheta2, 0.])

@try2jit
def _transform(s: np.ndarray) -> np.ndarray:
    theta1, theta2, dtheta1, dtheta2 = s

    return np.array([np.cos(theta1), np.sin(theta1), np.cos(theta2), np.sin(theta2), dtheta1, dtheta2])

@try2jit
def _isTerminal(s: np.ndarray) -> bool:
    return -np.cos(s[0]) - np.cos(s[1] + s[0]) > 1.

class Acrobot(BaseEnvironment):
    # -------------
    # -- Physics --
    # -------------
    physical_constants = {
        'gravity': 9.8,
        'link1_length': 1.,
        'link1_mass': 1.,
        'link1_com': 0.5,  # center of mass
        # link2 length is fixed
        'link2_mass': 1.,
        'link2_com': 0.5
    }

    per_step_constants = {
        'dt': DeltaDist(0.2),
        'force': DeltaDist(1.0),
    }

    randomized_constants = {
        'gravity': ClippedGaussian(mean=9.8, stddev=2.0, mi=5.0, ma=13.8),
        'link1_length': Uniform(mi=0.75, ma=1.25),
        'link1_mass': Uniform(mi=0.75, ma=1.25),
        'link1_com': ClippedGaussian(mean=0.5, stddev=0.1, mi=0.3, ma=0.7),
        # link2 length is fixed
        'link2_mass': Uniform(mi=0.75, ma=1.25),
        'link2_com': ClippedGaussian(mean=0.5, stddev=0.1, mi=0.3, ma=0.7),
    }

    per_step_random_constants = {
        # use clipped gaussian to enforce a lower-bound constraint on how fast we can sample
        # realistically, we can see very long delays but we can never sample faster than the equipment allows
        'dt': 0.99 * ClippedGaussian(mean=0.2, stddev=0.02, mi=0.125) + 0.01 * Gamma(shape=0.1, scale=2.0),

        # note this isn't clipped, force can flip signs with low probability
        'force': Gaussian(mean=1.0, stddev=0.4),
    }

    def __init__(self, randomize: bool = False, seed: int = 0):
        super().__init__(seed)
        self.randomize = randomize
        self._state = np.zeros(4)

        self.start_rng = np.random.default_rng(seed)

        if randomize:
            self.physical_constants = sampleChildren(self.randomized_constants, self.rng)
            self.per_step_constants = self.per_step_random_constants

        self._dsdt = partial(_dsdt,
            self.physical_constants['link1_length'],
            self.physical_constants['link1_mass'],
            self.physical_constants['link2_mass'],
            self.physical_constants['link1_com'],
            self.physical_constants['link2_com'],
            self.physical_constants['gravity'],
        )

    # -------------------------
    # -- Dynamics equations --
    # -------------------------

    def nextState(self, s: np.ndarray, a: float):
        dt = self.per_step_constants['dt'].sample(self.rng)
        force = self.per_step_constants['force'].sample(self.rng)

        a = (a - 1.) * force

        sa = np.append(s, a)
        spa = rungeKutta(self._dsdt, sa, np.array([0, dt]))

        # only need the last result of the integration
        spa = spa[-1]
        sp = spa[:-1]

        ma_vel1 = 4 * np.pi
        ma_vel2 = 9 * np.pi

        sp[0] = wrap(sp[0], -np.pi, np.pi)
        sp[1] = wrap(sp[1], -np.pi, np.pi)
        sp[2] = clip(sp[2], -ma_vel1, ma_vel1)
        sp[3] = clip(sp[3], -ma_vel2, ma_vel2)

        return sp

    def actions(self, s: np.ndarray):
        return [0, 1, 2]

    def reward(self, s: np.ndarray, a: float, sp: np.ndarray):
        return -1. if not self.terminal(s, a, sp) else 0.

    def terminal(self, s: np.ndarray, a: float, sp: np.ndarray):
        return _isTerminal(sp)

    # ------------------------
    # -- Stateful functions --
    # ------------------------

    def start(self):
        start = self.start_rng.uniform(-.1, .1, size=4)
        self._state = start

        return _transform(start)

    def step(self, action: float):
        sp = self.nextState(self._state, action)
        r = self.reward(self._state, action, sp)
        t = self.terminal(self._state, action, sp)

        self._state = sp

        return (r, _transform(sp), t)

    def setState(self, state: np.ndarray):
        self._state = state.copy()

    def copy(self):
        m = Acrobot(randomize=self.randomize, seed=self._seed)
        m._state = self._state.copy()
        m.physical_constants = self.physical_constants
        m.per_step_constants = self.per_step_constants

        # copy derivative function because state variables changed
        m._dsdt = self._dsdt

        return m

class StochasticAcrobot(Acrobot):
    def __init__(self, seed: int = 0):
        super().__init__(randomize=True, seed=seed)

addToCategory('classic-control', Acrobot)
addToCategory('stochastic', StochasticAcrobot)



import numpy as np
from typing import Callable, Sequence, List
from PyRlEnvs.FiniteDynamics import FiniteDynamics
from .Elements import Element
from .utils import Coords, findFirstTrigger, getState, getCoords

# --------------
# Static builder
# --------------

# To make sure we can build gridworlds statically
# use a builder class to collect all of the meta-data
class GridWorldBuilder:
    def __init__(self, shape: Coords):
        self.shape = shape
        self.costToGoal = True

        self.elements: List[Element] = []

    def addElement(self, element: Element):
        element.init(self.shape)
        self.elements.append(element)

    def addElements(self, elements: Sequence[Element]):
        for element in elements:
            self.addElement(element)

    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        for element in self.elements:
            element.apply(d0, K, T, R)

    def build(self):
        return buildGridWorld(self)


# ----------------------
# Common gridworld logic
# ----------------------

# To make sure the factory function does not get too complex
# pull shared logic out into its own base class
class _BaseGridWorld(FiniteDynamics):
    shape: Coords
    elements: List[Element]

    getState: Callable[[Coords], int]
    getCoords: Callable[[int], Coords]

    def addElement(self, element: Element):
        element.init(self.shape)
        element.apply(self.d0, self.K, self.T, self.Rs)
        self.elements.append(element)

    def addElements(self, elements: List[Element]):
        for element in elements:
            self.addElement(element)

    @classmethod
    def show(cls):
        width, height = cls.shape
        row_str = '+---' * (width) + '+'
        print(row_str)
        for y in range(height - 1, -1, -1):
            print('|', end='')
            for x in range(width):
                s = cls.getState((x, y))
                element = findFirstTrigger(cls.elements, s, 0, s)
                if element and element.name:
                    print(f' {element.name[0]} ', end='')
                else:
                    print('   ', end='')

                print('|', end='')

            print()
            print(row_str)

# generate the basic transition and reward kernels for an empty gridworld
# we can later modify these when we add new elements
def _buildKernels(shape: Coords, costToGoal: bool):
    width, height = shape
    states = width * height

    def _getState(coords: Coords):
        return getState(coords, shape)

    _K = np.zeros((states, 4, states))
    _R = np.zeros((states, 4, states))

    for x in range(width):
        for y in range(height):
            s = _getState((x, y))
            for a in range(4):
                # UP
                if a == 0:
                    sp = _getState((x, y + 1))
                # RIGHT
                elif a == 1:
                    sp = _getState((x + 1, y))
                # DOWN
                elif a == 2:
                    sp = _getState((x, y - 1))
                # LEFT
                else:
                    sp = _getState((x - 1, y))

                _K[s, a, sp] = 1

                if costToGoal:
                    _R[s, a, sp] = -1

    return _K, _R

# --------------------------
# Gridworld Factory Function
# --------------------------
def buildGridWorld(builder: GridWorldBuilder):
    width, height = builder.shape
    states = width * height
    actions = 4

    # partially apply the getState function
    # to simplify code a bit
    def _getState(coords: Coords):
        return getState(coords, builder.shape)

    def _getCoords(state: int):
        return getCoords(state, builder.shape)

    # build the dynamics tensors
    _T = np.zeros((states, actions, states))
    _d0 = np.zeros(states)

    _K, _R = _buildKernels(builder.shape, builder.costToGoal)

    # modify the tensors with individual elements
    builder.apply(_d0, _K, _T, _R)

    # TODO: consider if this should be a warning
    if _d0.sum() == 0:
        _d0[0] = 1.0

    # ensure this ends up as a probability distribution
    _d0 = _d0 / _d0.sum()

    class GridWorld(_BaseGridWorld):
        shape = builder.shape
        elements = builder.elements

        num_states = states
        num_actions = actions

        K = _K
        Rs = _R
        T = _T
        d0 = _d0

        getState = _getState
        getCoords = _getCoords

    return GridWorld



import numpy as np
from PyRlEnvs.utils.math import try2FastJit
import PyRlEnvs.utils.random as Random
from typing import Any, List, Set
from PyRlEnvs.FiniteDynamics import FiniteDynamics
from PyRlEnvs.domains.GridWorld.utils import Coords, getCoords, getState

# -------------------
# Maze generating alg
# -------------------

def sample(_shape: Coords, costToGoal: bool = True, seed: int = 0):
    rng = np.random.RandomState(seed)

    # collect some metadata
    width, height = _shape
    states = width * height

    # build the environment transition kernels
    _K = np.zeros((states, 4, states))
    _R = np.zeros((states, 4, states))
    _T = np.zeros((states, 4, states))
    _d0 = np.zeros(states)

    # build the set of states that the wall-builder
    # has not yet visited, initially this is every state
    unvisited = set(range(states))

    # pick a state and mark it as visited
    # we will guarantee that all paths connect to this state
    # at some point
    start = Random.choice(unvisited, rng)
    unvisited.remove(start)

    # the terminal state is in the top right of the maze
    terminal_state = getState((width - 1, height - 1), _shape)
    _T[terminal_state, :, terminal_state] = 1

    # build a simple -1 per step reward function
    r = -1 if costToGoal else 0
    rt = -1 if costToGoal else 1

    # sample paths until we've visited every state
    while len(unvisited) > 0:
        path = _samplePath(unvisited, _shape, rng)

        # walk the path and "activate" every transition from
        # prev -> cell. need to carefully handle termination states
        # also make sure the agent can walk backwards through the space
        # by connecting cell -> prev
        prev = None
        for cell in path:
            if prev is not None:
                unvisited.remove(prev)

                # mark all available actions from prev -> cell
                # might be multiple actions due to bumping into walls
                for a in actions(prev, cell, _shape):
                    _K[prev, a, cell] = 1
                    _R[prev, a, cell] = r

                    if cell == terminal_state:
                        _T[prev, a, cell] = 1
                        _R[prev, a, cell] = rt

                # mark all available actions from cell -> prev
                for a in actions(cell, prev, _shape):
                    _K[cell, a, prev] = 1
                    _R[cell, a, prev] = r

                    if prev == terminal_state:
                        _T[cell, a, prev] = 1
                        _R[cell, a, prev] = rt

            prev = cell

    # now we need to make sure all self-connections exist
    # that is, if I run into a wall then I stay in the same state
    for state in range(states):
        for a in range(4):
            # if this action doesn't lead anywhere, then it needs to be a self-transition
            if _K[state, a].sum() == 0:
                _K[state, a, state] = 1
                _R[state, a, state] = r

    # set start state as the bottom left
    start = getState((0, 0), _shape)
    _d0[start] = 1

    class WilsonMaze(_WilsonMaze):
        shape = _shape

        num_states = states
        num_actions = 4

        K = _K
        Rs = _R
        T = _T
        d0 = _d0

    return WilsonMaze

# ------------------------
# Internal utility methods
# ------------------------

class _WilsonMaze(FiniteDynamics):
    shape: Coords

    @classmethod
    def getState(cls, coords: Coords):
        return getState(coords, cls.shape)

    @classmethod
    def getCoords(cls, state: int):
        return getCoords(state, cls.shape)

    @classmethod
    def show(cls):
        width, height = cls.shape
        tops: List[List[str]] = []
        sides: List[List[str]] = []

        for y in range(height):
            top: List[str] = []
            side: List[str] = []
            for x in range(width):
                state = cls.getState((x, y))

                if cls.K[state, 0, state] == 0:
                    top.append(' ')
                else:
                    top.append('-')

                if cls.K[state, 3, state] == 0:
                    side.append(' ')
                else:
                    side.append('|')

            top.append('|')
            side.append('|')

            tops.append(top)
            sides.append(side)

        for y in range(height - 1, -1, -1):
            for x in range(width):
                print('+' + tops[y][x] * 3, end='')
            print('+')

            for x in range(width):
                print(f'{sides[y][x]}   ', end='')
            print('|')

        print('+---' * width + '+')

@try2FastJit
def neighbors(state: int, shape: Coords):
    x, y = getCoords(state, shape)

    return set([
        getState((x + 1, y), shape),
        getState((x - 1, y), shape),
        getState((x, y + 1), shape),
        getState((x, y - 1), shape),
    ])

@try2FastJit
def actions(state: int, next_state: int, shape: Coords):
    x, y = getCoords(state, shape)

    # trick numba into know the type of this empty list
    ret: List[int] = [i for i in range(0)]

    up = getState((x, y + 1), shape)
    if up == next_state:
        ret.append(0)

    right = getState((x + 1, y), shape)
    if right == next_state:
        ret.append(1)

    down = getState((x, y - 1), shape)
    if down == next_state:
        ret.append(2)

    left = getState((x - 1, y), shape)
    if left == next_state:
        ret.append(3)

    return ret

def _samplePath(unvisited: Set[int], shape: Coords, rng: Any):
    cell = Random.choice(unvisited, rng)

    path = [cell]

    while cell in unvisited:
        cell = Random.choice(neighbors(cell, shape), rng)

        # delete loops, otherwise we cannot guarantee solvability
        if cell in path:
            path = path[0:path.index(cell) + 1]
        else:
            path.append(cell)

    return path

# -----------------------
# Provide a default class
# -----------------------
Maze10x10 = sample((10, 10))
Maze5x5 = sample((5, 5))



import numpy as np
from PyRlEnvs.Category import addToCategory
from PyRlEnvs.domains.GridWorld.Elements import Element, WallState
from PyRlEnvs.domains.GridWorld import GridWorldBuilder

class UniformStart(Element):
    def apply(self, d0: np.ndarray, K: np.ndarray, T: np.ndarray, R: np.ndarray):
        for s in range(K.shape[0]):
            d0[s] = 1

def build():
    fourRoomsBuilder = GridWorldBuilder((11, 11))
    fourRoomsBuilder.addElement(UniformStart())
    fourRoomsBuilder.addElements([
        # Split top/bottom left
        WallState((0, 5)),
        WallState((2, 5)),
        WallState((3, 5)),
        WallState((4, 5)),
        WallState((5, 5)),
        # Split top left/right
        WallState((5, 10)),
        WallState((5, 9)),
        WallState((5, 7)),
        WallState((5, 6)),
        # Split bottom left/right
        WallState((5, 4)),
        WallState((5, 3)),
        WallState((5, 2)),
        WallState((5, 0)),
        # Split top/bottom right
        WallState((6, 4)),
        WallState((7, 4)),
        WallState((9, 4)),
        WallState((10, 4)),
    ])

    return fourRoomsBuilder.build()

FourRooms = build()

addToCategory('gridworld', FourRooms)
addToCategory('finite-dynamics', FourRooms)



"""
TODO:
From <cite Kolter>
"""

import numpy as np
from typing import Optional
from PyRlEnvs.utils.math import immutable
from PyRlEnvs.Category import addToCategory
from PyRlEnvs.FiniteDynamics import FiniteDynamics

S1 = 0
S2 = 1

def _buildTransition():
    K = np.zeros((2, 2, 2))
    K[S1, 0, S1] = 1
    K[S1, 1, S2] = 1
    K[S2, 1, S2] = 1
    K[S2, 0, S1] = 1

    return K

# there are actually infinitely many reward functions that result in the counterexample
# this one is a fairly literal translation from the paper
def _buildReward():
    Rs = np.zeros((2, 2, 2))
    Rs[S1, 0, S1] = -0.01475
    Rs[S1, 1, S2] = -0.01475
    Rs[S2, 1, S2] = 0.03525
    Rs[S2, 0, S1] = 0.03525

    return Rs

class KolterCounterexample(FiniteDynamics):
    num_states = 2
    num_actions = 2

    K = immutable(_buildTransition())
    Rs = immutable(_buildReward())
    T = immutable(np.zeros((2, 2, 2)))
    d0 = immutable(np.array([1., 0]))

# some utility functions to encode other important parts of the problem spec
# not necessarily environment specific, but this is as good a place as any to store them
def buildProblem(eps: float = 0.01, p: Optional[float] = None):
    # if a probability is not specified, then use the counterexample probability
    if p is None:
        p = (2961 + 45240 * eps + 40400 * eps**2) / (4141 + 84840 * eps + 40400 * eps**2)

    mu = immutable(np.array([p, 1 - p]))
    def behavior(s: int):
        return mu

    pi = immutable(np.array([0.5, 0.5]))
    def target(s: int):
        return pi

    X = immutable(np.array([
        [1],
        [1.05 + eps],
    ]))

    gamma = 0.99

    return behavior, target, X, gamma

addToCategory('ope', KolterCounterexample)
addToCategory('ope-counterexample', KolterCounterexample)
addToCategory('finite-dynamics', KolterCounterexample)



import numpy as np
from typing import Callable

def rungeKutta(derivs: Callable[[np.ndarray, float], np.ndarray], y0: np.ndarray, ts: np.ndarray):
    yout = np.zeros((len(ts), len(y0)), dtype=y0.dtype)

    yout[0] = y0

    for i in range(len(ts) - 1):
        t = ts[i]
        dt = ts[i + 1] - t
        y0 = yout[i]

        k1 = derivs(y0, t)
        k2 = derivs(y0 + (dt / 2.0) * k1, t + (dt / 2.0))
        k3 = derivs(y0 + (dt / 2.0) * k2, t + (dt / 2.0))
        k4 = derivs(y0 + dt * k3, t + dt)

        yout[i + 1] = y0 + (dt / 6.0) * (k1 + 2 * k2 + 2 * k3 + k4)

    return yout

def euler(derivs: Callable[[np.ndarray, float], np.ndarray], y0: np.ndarray, ts: np.ndarray):
    yout = np.zeros((len(ts), len(y0)), dtype=y0.dtype)

    yout[0] = y0

    for i in range(len(ts) - 1):
        t = ts[i]
        dt = ts[i + 1] - t

        y0 = yout[i]
        ydot = derivs(y0, t)

        yout[i + 1] = y0 + ydot * dt

    return yout



"""
From Sutton and Barto 2018
Example 11.4, the MRP that shows that the bellman error is "not learnable"

This implementation is set up as an MDP. We recover the original MRP if the behavior policy is uniform random.
"""

from PyRlEnvs.utils.math import immutable
from PyRlEnvs.Category import addToCategory
import numpy as np
from PyRlEnvs.FiniteDynamics import FiniteDynamics

LEFT = 0
RIGHT = 1

A = 0
B = 1
Bp = 2

def _buildTransitionKernel():
    K = np.zeros((3, 2, 3))
    K[A, LEFT, B] = 1
    K[A, RIGHT, Bp] = 1
    K[B, :, A] = 1
    K[Bp, LEFT, B] = 1
    K[Bp, RIGHT, Bp] = 1

    return K

def _buildRewardKernel():
    R = np.zeros((3, 2, 3))
    R[B, :, A] = 1
    R[Bp, LEFT, B] = -1
    R[Bp, RIGHT, Bp] = -1

    return R

class BECounterexample(FiniteDynamics):
    num_states = 3
    num_actions = 2

    K = immutable(_buildTransitionKernel())
    Rs = immutable(_buildRewardKernel())

    T = immutable(np.zeros((3, 2, 3)))
    d0 = immutable(np.array([1., 0, 0]))

# some utility functions to encode other important parts of the problem spec
# not necessarily environment specific, but this is as good a place as any to store them
def behaviorPolicy(s: int):
    return np.array([0.5, 0.5])

def representationMatrix():
    return np.array([
        [1., 0],
        [0, 1],
        [0, 1],
    ])

addToCategory('ope', BECounterexample)
addToCategory('ope-counterexample', BECounterexample)
addToCategory('finite-dynamics', BECounterexample)



from glob import glob
import os

import PyRlEnvs.Category as Cat

# categories are stored along with the modules
# we don't really want to load all of the modules whenever the library is loaded,
# so we'll make that an opt-in feature
_categories_built = False
def _buildCategories():
    global _categories_built

    if _categories_built:
        return

    _categories_built = True

    domain_paths = glob(os.path.join(os.path.dirname(__file__), 'domains', '*.py'))
    domain_paths = (path for path in domain_paths if '__init__.py' not in path)
    domain_paths = (path for path in domain_paths if os.path.isfile(path))

    for path in domain_paths:
        __import__(path)


def getCategory(category: str):
    _buildCategories()

    return Cat.getCategory(category)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

print(find_packages(exclude=["tests*", "scripts*"]))

setup(
    name='PyRlEnvs-andnp',
    url='https://github.com/andnp/PyRlEnvs.git',
    author='Andy Patterson',
    author_email='andnpatterson@gmail.com',
    packages=find_packages(exclude=['tests*', 'scripts*']),
    package_data={"PyRlEnvs": ["py.typed"]},
    version="1.1.0",
    license="MIT",
    description="A handful of fast environments for running RL experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    install_requires=["numpy>=1.19.5", "numba>=0.52.0", "scipy>=1.5.4", "rlglue-andnp"],
    extras_require={
        "dev": [
            "build",
            "commitizen",
            "mypy>=0.942",
            "flake8>=4.0.1",
            "gym>=0.18.0",
            "pre-commit",
            "pygame",
            "pipenv-setup[black]",
            "twine",
        ],
    },
)



import numpy as np
from functools import partial
from PyRlEnvs.utils.distributions import ClippedGaussian, DeltaDist, Gamma, Gaussian, Uniform, sampleChildren
from PyRlEnvs.utils.math import clip, try2jit
from PyRlEnvs.utils.numerical import euler
from PyRlEnvs.Category import addToCategory
from PyRlEnvs.BaseEnvironment import BaseEnvironment

@try2jit
def _dsdt(g: float, m: float, k: float, sa: np.ndarray, t: float):
    p, v, f = sa

    # compute derivative of velocity (acc)
    dv = -g * np.cos(3 * p) + (f / m) - k * v

    return np.array([v, dv, 0.])

# use the hand-crafted (incorrect) euler update found in OpenAI gym and the book
# note that timesteps don't really make sense here, we increase acceleration
# and see an instantaneous change in velocity *and* position afterwards, should
# only see change in velocity at this time, then change in position at *next* time
# keep this around for consistency though
@try2jit
def _nextState(s: np.ndarray, a: int):
    a = a - 1
    p: float = s[0]
    v: float = s[1]

    v = v + 0.001 * a - 0.0025 * np.cos(3 * p)

    if v < -0.07:
        v = -0.07
    elif v >= 0.07:
        v = 0.07

    p += v

    if p < -1.2:
        p = -1.2
        v = 0

    return np.array([p, v])

class MountainCar(BaseEnvironment):
    # -------------
    # -- Physics --
    # -------------
    physical_constants = {
        'gravity': 0.0025,
        'cart_mass': 10,
        'friction': 0.0,
    }

    per_step_constants = {
        'dt': DeltaDist(0.1),
        'force': DeltaDist(.1),
    }

    randomized_constants = {
        # keep stddev low to prevent overly powerful gravity simulations
        # which can make the problem unsolvable, better to err on easier than impossible
        'gravity': ClippedGaussian(mean=0.0025, stddev=0.001, mi=0.001, ma=0.0035),
        'cart_mass': Uniform(mi=8.5, ma=11.5),
        'friction': Uniform(mi=0.0, ma=0.3),
    }

    per_step_random_constants = {
        # use clipped gaussian to enforce a lower-bound constraint on how fast we can sample
        # realistically, we can see very long delays but we can never sample faster than the equipment allows
        'dt': 0.99 * ClippedGaussian(mean=0.1, stddev=0.02, mi=0.05) + 0.01 * Gamma(shape=0.1, scale=1.0),

        # note this isn't clipped, force can flip signs with low probability
        # we use a fairly high stddev to help ensure solvability
        'force': Gaussian(mean=0.1, stddev=0.02),
    }

    def __init__(self, randomize: bool = False, seed: int = 0):
        super().__init__(seed)
        self.randomize = randomize
        self._state = np.array([0, 0])

        self.start_rng = np.random.default_rng(seed)

        if randomize:
            self.physical_constants = sampleChildren(self.randomized_constants, self.rng)
            self.per_step_constants = self.per_step_random_constants

        self._dsdt = partial(_dsdt,
            self.physical_constants['gravity'],
            self.physical_constants['cart_mass'],
            self.physical_constants['friction'],
        )

    def _integrate(self, s: np.ndarray, force: float):
        dt = self.per_step_constants['dt'].sample(self.rng)

        sa = np.append(s, force)
        spa = euler(self._dsdt, sa, np.array([0, dt]))

        # only need last result of integration
        spa = spa[-1]
        sp = spa[:-1]

        # enforce bounds outside of the integrator
        sp[1] = clip(sp[1], -0.07, 0.07)
        if sp[0] < -1.2:
            sp[0] = -1.2
            sp[1] = 0.

        return sp

    def nextState(self, s: np.ndarray, a: int):
        # convert a discrete action into a continuous force
        force = self.per_step_constants['force'].sample(self.rng) * (a - 1)

        return self._integrate(s, force)

    def actions(self, s: np.ndarray):
        return [0, 1, 2]

    def reward(self, s: np.ndarray, a: int, sp: np.ndarray):
        return -1

    def terminal(self, s: np.ndarray, a: int, sp: np.ndarray):
        p, _ = sp

        return p >= 0.5

    def start(self):
        position = -0.6 + self.start_rng.random() * 0.2
        velocity = 0

        start = np.array([position, velocity])
        self._state = start

        return start

    def step(self, action: int):
        sp = self.nextState(self._state, action)
        r = self.reward(self._state, action, sp)
        t = self.terminal(self._state, action, sp)

        self._state = sp

        return (r, sp.copy(), t)

    def setState(self, state: np.ndarray):
        self._state = state.copy()

    def copy(self):
        m = MountainCar(randomize=self.randomize, seed=self._seed)
        m._state = self._state.copy()
        m.physical_constants = self.physical_constants
        m.per_step_constants = self.per_step_constants

        m._dsdt = self._dsdt

        return m

class GymMountainCar(MountainCar):
    def __init__(self, seed: int = 0):
        super().__init__(randomize=False, seed=seed)

    def nextState(self, s: np.ndarray, a: int):
        return _nextState(s, a)

class StochasticMountainCar(MountainCar):
    def __init__(self, seed: int = 0):
        super().__init__(randomize=True, seed=seed)

class ContinuousActionMountainCar(MountainCar):
    def nextState(self, s: np.ndarray, a: float):
        force = np.clip(a, -3, 3)

        return self._integrate(s, force)


addToCategory('classic-control', MountainCar)
addToCategory('sutton-barto', GymMountainCar)
addToCategory('stochastic', StochasticMountainCar)



"""
TODO:
From <cite Baird>
"""

from PyRlEnvs.utils.math import immutable
from PyRlEnvs.Category import addToCategory
import numpy as np
from PyRlEnvs.FiniteDynamics import FiniteDynamics

OUT = 0
IN = 1


def _buildTransitionKernel():
    K = np.zeros((7, 2, 7))
    K[:, OUT, :6] = 1 / 6
    K[:, IN, 6] = 1

    return K

class BairdCounterexample(FiniteDynamics):
    num_states = 7
    num_actions = 2

    K = immutable(_buildTransitionKernel())
    Rs = immutable(np.zeros((7, 2, 7)))

    T = immutable(np.zeros((7, 2, 7)))
    d0 = immutable(np.array([0, 0, 0, 0, 0, 0, 1.]))

# some utility functions to encode other important parts of the problem spec
# not necessarily environment specific, but this is as good a place as any to store them
def behaviorPolicy(s: int):
    return np.array([6 / 7, 1 / 7])

def targetPolicy(s: int):
    return np.array([0, 1.])

def representationMatrix():
    return np.array([
        [1, 2, 0, 0, 0, 0, 0, 0],
        [1, 0, 2, 0, 0, 0, 0, 0],
        [1, 0, 0, 2, 0, 0, 0, 0],
        [1, 0, 0, 0, 2, 0, 0, 0],
        [1, 0, 0, 0, 0, 2, 0, 0],
        [1, 0, 0, 0, 0, 0, 2, 0],
        [2, 0, 0, 0, 0, 0, 0, 1.],
    ])

def initialWeights():
    return np.array([1, 1, 1, 1, 1, 1, 1, 10.])

addToCategory('ope', BairdCounterexample)
addToCategory('ope-counterexample', BairdCounterexample)
addToCategory('finite-dynamics', BairdCounterexample)
